# Copyright 2021 PingCAP, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# See the License for the specific language governing permissions and
# limitations under the License.

from django.db.backends.mysql.schema import (
    DatabaseSchemaEditor as MysqlDatabaseSchemaEditor,
)


class DatabaseSchemaEditor(MysqlDatabaseSchemaEditor):
    # Unsupported add column and foreign key in single statement
    # https://github.com/pingcap/tidb/issues/45474
    sql_create_column_inline_fk = None

    @property
    def sql_delete_check(self):
        return "ALTER TABLE %(table)s DROP CHECK %(name)s"

    @property
    def sql_rename_column(self):
        return "ALTER TABLE %(table)s CHANGE %(old_column)s %(new_column)s %(type)s"

    def skip_default_on_alter(self, field):
        if self._is_limited_data_type(field):
            # TiDB doesn't support defaults for BLOB/TEXT/JSON in the
            # ALTER COLUMN statement.
            return True
        return False

    @property
    def _supports_limited_data_type_defaults(self):
        return False

    def _field_should_be_indexed(self, model, field):
        if not field.db_index or field.unique:
            return False
        # No need to create an index for ForeignKey fields except if
        # db_constraint=False because the index from that constraint won't be
        # created.
        if field.get_internal_type() == "ForeignKey" and field.db_constraint:
            return False
        return not self._is_limited_data_type(field)

    def add_field(self, model, field):
        if field.unique:
            # TiDB does not support multiple operations with a single DDL statement,
            # so we need to execute the unique constraint creation separately.
            field._unique = False
            super().add_field(model, field)
            field._unique = True
            self.execute(self._create_unique_sql(model, [field]))
        else:
            super().add_field(model, field)

    def table_sql(self, model):
        sql, params = super().table_sql(model)
        tidb_auto_id_cache = getattr(model._meta, "tidb_auto_id_cache", None)
        if tidb_auto_id_cache is not None:
            sql += " AUTO_ID_CACHE %s" % tidb_auto_id_cache
        return sql, params



import json

from django.test import TestCase
from django.test.utils import CaptureQueriesContext
from django.db import connection, transaction

from .models import Course


class TiDBExplainTests(TestCase):
    SUPPORTED_FORMATS = {"TRADITIONAL", "ROW", "BRIEF", "DOT", "TIDB_JSON"}

    def test_explain_with_supported_format(self):
        for format in self.SUPPORTED_FORMATS:
            with self.subTest(format=format), transaction.atomic():
                with CaptureQueriesContext(connection) as captured_queries:
                    result = Course.objects.filter(name="test").explain(format=format)
                    self.assertTrue(
                        captured_queries[0]["sql"].startswith(
                            connection.ops.explain_prefix + f' FORMAT="{format}"'
                        )
                    )
                    if format == "TIDB_JSON":
                        try:
                            json.loads(result)
                        except json.JSONDecodeError as e:
                            self.fail(
                                f"QuerySet.explain() result is not valid JSON: {e}"
                            )

    def test_explain_analyze_with_supported_format(self):
        for format in self.SUPPORTED_FORMATS:
            with self.subTest(format=format), transaction.atomic():
                with CaptureQueriesContext(connection) as captured_queries:
                    result = Course.objects.filter(name="test").explain(
                        analyze=True, format=format
                    )
                    self.assertTrue(
                        captured_queries[0]["sql"].startswith(
                            connection.ops.explain_prefix
                            + f' ANALYZE FORMAT="{format}"'
                        )
                    )
                    if format == "TIDB_JSON":
                        try:
                            json.loads(result)
                        except json.JSONDecodeError as e:
                            self.fail(
                                f"QuerySet.explain() result is not valid JSON: {e}"
                            )

    def test_explain_with_unsupported_format(self):
        with self.assertRaises(ValueError):
            Course.objects.filter(name="test").explain(format="JSON")

    def test_explain_analyze_with_unsupported_format(self):
        with self.assertRaises(ValueError):
            Course.objects.filter(name="test").explain(analyze=True, format="JSON")

    def test_explain_with_unkonwn_option(self):
        with self.assertRaises(ValueError):
            Course.objects.filter(name="test").explain(unknown_option=True)

    def test_explain_with_default_params(self):
        with transaction.atomic():
            with CaptureQueriesContext(connection) as captured_queries:
                Course.objects.filter(name="test").explain()
                self.assertTrue(
                    captured_queries[0]["sql"].startswith(
                        connection.ops.explain_prefix + ' FORMAT="ROW"'
                    )
                )



from django.test import TransactionTestCase
from django.test.utils import isolate_apps
from django.db import models, connection


class TiDBDDLTests(TransactionTestCase):
    available_apps = ["tidb"]

    def get_indexes(self, table):
        """
        Get the indexes on the table using a new cursor.
        """
        with connection.cursor() as cursor:
            return [
                c["columns"][0]
                for c in connection.introspection.get_constraints(
                    cursor, table
                ).values()
                if c["index"] and len(c["columns"]) == 1
            ]

    def get_uniques(self, table):
        with connection.cursor() as cursor:
            return [
                c["columns"][0]
                for c in connection.introspection.get_constraints(
                    cursor, table
                ).values()
                if c["unique"] and len(c["columns"]) == 1
            ]

    @isolate_apps("tidb")
    def test_should_create_db_index(self):
        # When define a model with db_index=True, TiDB should create a db index
        class Tag(models.Model):
            title = models.CharField(max_length=255, db_index=True)

            class Meta:
                app_label = "tidb"

        with connection.schema_editor() as editor:
            editor.create_model(Tag)
        self.assertIn("title", self.get_indexes("tidb_tag"))

        new_field = models.CharField(max_length=255, db_index=True)
        new_field.set_attributes_from_name("new_field")
        with connection.schema_editor() as editor:
            editor.add_field(Tag, new_field)
        self.assertIn("new_field", self.get_indexes("tidb_tag"))

    @isolate_apps("tidb")
    def test_should_create_db_index_for_foreign_key_with_no_db_constraint(self):
        # When define a model with ForeignKey, TiDB should not create a db index
        class Node1(models.Model):
            title = models.CharField(max_length=255)

            class Meta:
                app_label = "tidb"

        class Node2(models.Model):
            node1 = models.ForeignKey(
                Node1, on_delete=models.CASCADE, db_constraint=False
            )

            class Meta:
                app_label = "tidb"

        with connection.schema_editor() as editor:
            editor.create_model(Node1)
            editor.create_model(Node2)

        self.assertIn("node1_id", self.get_indexes("tidb_node2"))

    @isolate_apps("tidb")
    def test_add_unique_field(self):
        # issue: https://github.com/pingcap/django-tidb/issues/48
        class Node3(models.Model):
            title = models.CharField(max_length=255, unique=True)

            class Meta:
                app_label = "tidb"

        with connection.schema_editor() as editor:
            editor.create_model(Node3)
        self.assertIn("title", self.get_uniques("tidb_node3"))

        new_field = models.CharField(max_length=255, unique=True)
        new_field.set_attributes_from_name("new_field")
        with connection.schema_editor() as editor:
            editor.add_field(Node3, new_field)
        self.assertIn("new_field", self.get_uniques("tidb_node3"))

        parent = models.OneToOneField(Node3, models.CASCADE)
        parent.set_attributes_from_name("parent")
        with connection.schema_editor() as editor:
            editor.add_field(Node3, parent)
        self.assertIn("parent_id", self.get_uniques("tidb_node3"))



from django.db.models import BigAutoField
from django.core import checks

__all__ = [
    "BigAutoRandomField",
]


class BigAutoRandomField(BigAutoField):
    def __init__(
        self,
        verbose_name=None,
        name=None,
        shard_bits=5,
        range=64,
        **kwargs,
    ):
        self.shard_bits, self.range = shard_bits, range
        super().__init__(verbose_name, name, **kwargs)

    def get_internal_type(self):
        return "BigAutoRandomField"

    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_range(),
            *self._check_shard_bits(),
        ]

    def _check_shard_bits(self):
        try:
            shard_bits = int(self.shard_bits)
            if shard_bits < 1 or shard_bits > 15:
                raise ValueError()
        except TypeError:
            return [
                checks.Error(
                    "BigAutoRandomField must define a 'shard_bits' attribute.",
                    obj=self,
                )
            ]
        except ValueError:
            return [
                checks.Error(
                    "BigAutoRandomField 'shard_bits' attribute must be an integer between 1 and 15.",
                    obj=self,
                )
            ]
        else:
            return []

    def _check_range(self):
        try:
            range = int(self.range)
            if range < 32 or range > 64:
                raise ValueError()
        except TypeError:
            return [
                checks.Error(
                    "BigAutoRandomField must define a 'range' attribute.",
                    obj=self,
                )
            ]
        except ValueError:
            return [
                checks.Error(
                    "BigAutoRandomField 'range' attribute must be an integer between 32 and 64.",
                    obj=self,
                )
            ]
        else:
            return []

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.shard_bits is not None:
            kwargs["shard_bits"] = self.shard_bits
        if self.range is not None:
            kwargs["range"] = self.range
        return name, path, args, kwargs

    def db_type(self, connection):
        data = self.db_type_parameters(connection)
        if connection.tidb_version < (6, 3):
            # TiDB < 6.3 doesn't support define AUTO_RANDOM with range
            data_type = "bigint AUTO_RANDOM(%(shard_bits)s)"
        else:
            data_type = "bigint AUTO_RANDOM(%(shard_bits)s, %(range)s)"
        return data_type % data



import re
from contextlib import contextmanager

from django.db import models, connection
from django.core import checks, validators
from django.core.exceptions import ValidationError
from django.test import TestCase, TransactionTestCase, override_settings
from django.test.utils import isolate_apps

from django_tidb.fields import BigAutoRandomField
from .models import BigAutoRandomModel, BigAutoRandomExplicitInsertModel


class TiDBBigAutoRandomFieldTests(TestCase):
    model = BigAutoRandomModel
    explicit_insert_model = BigAutoRandomExplicitInsertModel
    documented_range = (-9223372036854775808, 9223372036854775807)
    rel_db_type_class = BigAutoRandomField

    @contextmanager
    def explicit_insert_allowed(self):
        with connection.cursor() as cursor:
            cursor.execute("SET @@allow_auto_random_explicit_insert=true")
            yield
            cursor.execute("SET @@allow_auto_random_explicit_insert=false")

    @property
    def backend_range(self):
        field = self.model._meta.get_field("value")
        internal_type = field.get_internal_type()
        return connection.ops.integer_field_range(internal_type)

    def test_documented_range(self):
        """
        Values within the documented safe range pass validation, and can be
        saved and retrieved without corruption.
        """
        min_value, max_value = self.documented_range

        with self.explicit_insert_allowed():
            instance = self.explicit_insert_model(value=min_value)
            instance.full_clean()
            instance.save()
            qs = self.explicit_insert_model.objects.filter(value__lte=min_value)
            self.assertEqual(qs.count(), 1)
            self.assertEqual(qs[0].value, min_value)

            instance = self.explicit_insert_model(value=max_value)
            instance.full_clean()
            instance.save()
            qs = self.explicit_insert_model.objects.filter(value__gte=max_value)
            self.assertEqual(qs.count(), 1)
            self.assertEqual(qs[0].value, max_value)

    def test_backend_range_save(self):
        """
        Backend specific ranges can be saved without corruption.
        """
        min_value, max_value = self.backend_range
        with self.explicit_insert_allowed():
            if min_value is not None:
                instance = self.explicit_insert_model(value=min_value)
                instance.full_clean()
                instance.save()
                qs = self.explicit_insert_model.objects.filter(value__lte=min_value)
                self.assertEqual(qs.count(), 1)
                self.assertEqual(qs[0].value, min_value)

            if max_value is not None:
                instance = self.explicit_insert_model(value=max_value)
                instance.full_clean()
                instance.save()
                qs = self.explicit_insert_model.objects.filter(value__gte=max_value)
                self.assertEqual(qs.count(), 1)
                self.assertEqual(qs[0].value, max_value)

    def test_backend_range_validation(self):
        """
        Backend specific ranges are enforced at the model validation level
        (#12030).
        """
        min_value, max_value = self.backend_range

        if min_value is not None:
            instance = self.model(value=min_value - 1)
            expected_message = validators.MinValueValidator.message % {
                "limit_value": min_value,
            }
            with self.assertRaisesMessage(ValidationError, expected_message):
                instance.full_clean()
            instance.value = min_value
            instance.full_clean()

        if max_value is not None:
            instance = self.model(value=max_value + 1)
            expected_message = validators.MaxValueValidator.message % {
                "limit_value": max_value,
            }
            with self.assertRaisesMessage(ValidationError, expected_message):
                instance.full_clean()
            instance.value = max_value
            instance.full_clean()

    def test_redundant_backend_range_validators(self):
        """
        If there are stricter validators than the ones from the database
        backend then the backend validators aren't added.
        """
        min_backend_value, max_backend_value = self.backend_range

        for callable_limit in (True, False):
            with self.subTest(callable_limit=callable_limit):
                if min_backend_value is not None:
                    min_custom_value = min_backend_value + 1
                    limit_value = (
                        (lambda: min_custom_value)
                        if callable_limit
                        else min_custom_value
                    )
                    ranged_value_field = self.model._meta.get_field("value").__class__(
                        validators=[validators.MinValueValidator(limit_value)]
                    )
                    field_range_message = validators.MinValueValidator.message % {
                        "limit_value": min_custom_value,
                    }
                    with self.assertRaisesMessage(
                        ValidationError, "[%r]" % field_range_message
                    ):
                        ranged_value_field.run_validators(min_backend_value - 1)

                if max_backend_value is not None:
                    max_custom_value = max_backend_value - 1
                    limit_value = (
                        (lambda: max_custom_value)
                        if callable_limit
                        else max_custom_value
                    )
                    ranged_value_field = self.model._meta.get_field("value").__class__(
                        validators=[validators.MaxValueValidator(limit_value)]
                    )
                    field_range_message = validators.MaxValueValidator.message % {
                        "limit_value": max_custom_value,
                    }
                    with self.assertRaisesMessage(
                        ValidationError, "[%r]" % field_range_message
                    ):
                        ranged_value_field.run_validators(max_backend_value + 1)

    def test_types(self):
        instance = self.model(tag="a")
        instance.save()
        self.assertIsInstance(instance.value, int)
        instance = self.model.objects.get()
        self.assertIsInstance(instance.value, int)

    def test_invalid_value(self):
        tests = [
            (TypeError, ()),
            (TypeError, []),
            (TypeError, {}),
            (TypeError, set()),
            (TypeError, object()),
            (TypeError, complex()),
            (ValueError, "non-numeric string"),
            (ValueError, b"non-numeric byte-string"),
        ]
        for exception, value in tests:
            with self.subTest(value):
                msg = "Field 'value' expected a number but got %r." % (value,)
                with self.assertRaisesMessage(exception, msg):
                    self.explicit_insert_model.objects.create(value=value)

    def test_rel_db_type(self):
        field = self.model._meta.get_field("value")
        rel_db_type = field.rel_db_type(connection)
        # Currently, We can't find a general way to get the auto_random info from the field.
        self.assertEqual(rel_db_type, "bigint")
        if connection.tidb_version < (6, 3):
            self.assertEqual(
                self.rel_db_type_class().db_type(connection), "bigint AUTO_RANDOM(5)"
            )
        else:
            self.assertEqual(
                self.rel_db_type_class().db_type(connection),
                "bigint AUTO_RANDOM(5, 64)",
            )


AUTO_RANDOM_PATTERN = re.compile(
    r"\/\*T!\[auto_rand\] AUTO_RANDOM\((\d+)(?:, (\d+))?\) \*\/"
)


class TiDBAutoRandomMigrateTests(TransactionTestCase):
    available_apps = ["tidb"]

    def get_primary_key(self, table):
        with connection.cursor() as cursor:
            primary_key_columns = connection.introspection.get_primary_key_columns(
                cursor, table
            )
            return primary_key_columns[0] if primary_key_columns else None

    def get_auto_random_info(self, table):
        # return (shard_bits, range)
        with connection.cursor() as cursor:
            cursor.execute(
                # It seems that SHOW CREATE TABLE is the only way to get the auto_random info.
                # Use parameterized query will add quotes to the table name, which will cause syntax error.
                f"SHOW CREATE TABLE {table}",
            )
            row = cursor.fetchone()
            if row is None:
                return (None, None)
            for line in row[1].splitlines():
                match = AUTO_RANDOM_PATTERN.search(line)
                if match:
                    return match.groups()
            return (None, None)

    @isolate_apps("tidb")
    @override_settings(DEFAULT_AUTO_FIELD="django_tidb.fields.BigAutoRandomField")
    def test_create_table_with_default_auto_field(self):
        class AutoRandomNode1(models.Model):
            title = models.CharField(max_length=255)

            class Meta:
                app_label = "tidb"

        with connection.schema_editor() as editor:
            editor.create_model(AutoRandomNode1)
        self.assertEqual(self.get_primary_key(AutoRandomNode1._meta.db_table), "id")
        self.assertIsInstance(AutoRandomNode1._meta.pk, BigAutoRandomField)
        self.assertEqual(
            self.get_auto_random_info(AutoRandomNode1._meta.db_table), ("5", None)
        )

    @isolate_apps("tidb")
    def test_create_table_explicit_auto_random_field(self):
        class AutoRandomNode2(models.Model):
            id = BigAutoRandomField(primary_key=True)
            title = models.CharField(max_length=255)

            class Meta:
                app_label = "tidb"

        with connection.schema_editor() as editor:
            editor.create_model(AutoRandomNode2)
        self.assertEqual(self.get_primary_key(AutoRandomNode2._meta.db_table), "id")
        self.assertIsInstance(AutoRandomNode2._meta.pk, BigAutoRandomField)
        self.assertEqual(
            self.get_auto_random_info(AutoRandomNode2._meta.db_table), ("5", None)
        )

    @isolate_apps("tidb")
    def test_create_table_explicit_auto_random_field_with_shard_bits(self):
        class AutoRandomNode3(models.Model):
            id = BigAutoRandomField(primary_key=True, shard_bits=10)
            title = models.CharField(max_length=255)

            class Meta:
                app_label = "tidb"

        with connection.schema_editor() as editor:
            editor.create_model(AutoRandomNode3)
        self.assertEqual(self.get_primary_key(AutoRandomNode3._meta.db_table), "id")
        self.assertIsInstance(AutoRandomNode3._meta.pk, BigAutoRandomField)
        self.assertEqual(
            self.get_auto_random_info(AutoRandomNode3._meta.db_table), ("10", None)
        )

    @isolate_apps("tidb")
    def test_create_table_explicit_auto_random_field_with_shard_bits_and_range(self):
        class AutoRandomNode4(models.Model):
            id = BigAutoRandomField(primary_key=True, shard_bits=10, range=60)
            title = models.CharField(max_length=255)

            class Meta:
                app_label = "tidb"

        with connection.schema_editor() as editor:
            editor.create_model(AutoRandomNode4)
        self.assertEqual(self.get_primary_key(AutoRandomNode4._meta.db_table), "id")
        self.assertIsInstance(AutoRandomNode4._meta.pk, BigAutoRandomField)
        self.assertEqual(
            self.get_auto_random_info(AutoRandomNode4._meta.db_table), ("10", "60")
        )

    @isolate_apps("tidb")
    def test_create_table_explicit_auto_random_field_with_range(self):
        class AutoRandomNode5(models.Model):
            id = BigAutoRandomField(primary_key=True, range=60)
            title = models.CharField(max_length=255)

            class Meta:
                app_label = "tidb"

        with connection.schema_editor() as editor:
            editor.create_model(AutoRandomNode5)
        self.assertEqual(self.get_primary_key(AutoRandomNode5._meta.db_table), "id")
        self.assertIsInstance(AutoRandomNode5._meta.pk, BigAutoRandomField)
        self.assertEqual(
            self.get_auto_random_info(AutoRandomNode5._meta.db_table), ("5", "60")
        )

    @isolate_apps("tidb")
    def test_create_table_explicit_auto_random_field_with_invalid_range(self):
        class AutoRandomNode6(models.Model):
            id = BigAutoRandomField(primary_key=True, range=31)

            class Meta:
                app_label = "tidb"

        id = AutoRandomNode6._meta.get_field("id")
        self.assertEqual(
            id.check(),
            [
                checks.Error(
                    "BigAutoRandomField 'range' attribute must be an integer between 32 and 64.",
                    obj=id,
                )
            ],
        )

        class AutoRandomNode7(models.Model):
            id = BigAutoRandomField(primary_key=True, range=None)

            class Meta:
                app_label = "tidb"

        id = AutoRandomNode7._meta.get_field("id")
        self.assertEqual(
            id.check(),
            [
                checks.Error(
                    "BigAutoRandomField must define a 'range' attribute.",
                    obj=id,
                )
            ],
        )

    @isolate_apps("tidb")
    def test_create_table_explicit_auto_random_field_with_invalid_shard_bits(self):
        class AutoRandomNode8(models.Model):
            id = BigAutoRandomField(primary_key=True, shard_bits=16)

            class Meta:
                app_label = "tidb"

        id = AutoRandomNode8._meta.get_field("id")
        self.assertEqual(
            id.check(),
            [
                checks.Error(
                    "BigAutoRandomField 'shard_bits' attribute must be an integer between 1 and 15.",
                    obj=id,
                )
            ],
        )

        class AutoRandomNode9(models.Model):
            id = BigAutoRandomField(primary_key=True, shard_bits=None)

            class Meta:
                app_label = "tidb"

        id = AutoRandomNode9._meta.get_field("id")
        self.assertEqual(
            id.check(),
            [
                checks.Error(
                    "BigAutoRandomField must define a 'shard_bits' attribute.",
                    obj=id,
                )
            ],
        )



import re

from django.db import models, connection
from django.db.utils import ProgrammingError
from django.test import TransactionTestCase
from django.test.utils import isolate_apps


AUTO_ID_CACHE_PATTERN = re.compile(r"\/\*T!\[auto_id_cache\] AUTO_ID_CACHE=(\d+) \*\/")


class TiDBAutoIDCacheTests(TransactionTestCase):
    available_apps = ["tidb"]

    def get_auto_id_cache_info(self, table):
        with connection.cursor() as cursor:
            cursor.execute(
                # It seems that SHOW CREATE TABLE is the only way to get the auto_random info.
                # Use parameterized query will add quotes to the table name, which will cause syntax error.
                f"SHOW CREATE TABLE {table}",
            )
            row = cursor.fetchone()
            if row is None:
                return None
            match = AUTO_ID_CACHE_PATTERN.search(row[1])
            if match:
                return match.groups()[0]
            return None

    @isolate_apps("tidb")
    def test_create_table_with_tidb_auto_id_cache_1(self):
        class AutoIDCacheNode1(models.Model):
            title = models.CharField(max_length=255)

            class Meta:
                app_label = "tidb"
                tidb_auto_id_cache = 1

        with connection.schema_editor() as editor:
            editor.create_model(AutoIDCacheNode1)
        self.assertEqual(
            self.get_auto_id_cache_info(AutoIDCacheNode1._meta.db_table), "1"
        )

    @isolate_apps("tidb")
    def test_create_table_with_tidb_auto_id_cache_non_1(self):
        class AutoIDCacheNode2(models.Model):
            title = models.CharField(max_length=255)

            class Meta:
                app_label = "tidb"
                tidb_auto_id_cache = 10

        with connection.schema_editor() as editor:
            editor.create_model(AutoIDCacheNode2)
        self.assertEqual(
            self.get_auto_id_cache_info(AutoIDCacheNode2._meta.db_table), "10"
        )

    @isolate_apps("tidb")
    def test_create_table_with_invalid_tidb_auto_id_cache(self):
        class AutoIDCacheNode3(models.Model):
            title = models.CharField(max_length=255)

            class Meta:
                app_label = "tidb"
                tidb_auto_id_cache = "invalid"

        with self.assertRaises(ProgrammingError):
            with connection.schema_editor() as editor:
                editor.create_model(AutoIDCacheNode3)

    @isolate_apps("tidb")
    def test_create_table_without_tidb_auto_id_cache(self):
        class AutoIDCacheNode4(models.Model):
            title = models.CharField(max_length=255)

            class Meta:
                app_label = "tidb"

        with connection.schema_editor() as editor:
            editor.create_model(AutoIDCacheNode4)
        self.assertIsNone(self.get_auto_id_cache_info(AutoIDCacheNode4._meta.db_table))



# Copyright 2021 PingCAP, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# See the License for the specific language governing permissions and
# limitations under the License.

from django.db.models.functions import Chr
from django.db.models import options
from django.db.migrations import state


def char(self, compiler, connection, **extra_context):
    # TiDB doesn't support utf16
    return self.as_sql(
        compiler,
        connection,
        function="CHAR",
        template="%(function)s(%(expressions)s USING utf8mb4)",
        **extra_context,
    )


def patch_model_functions():
    Chr.as_mysql = char


def patch_model_options():
    # Patch `tidb_auto_id_cache` to options.DEFAULT_NAMES,
    # so that user can define it in model's Meta class.
    options.DEFAULT_NAMES += ("tidb_auto_id_cache",)
    # Because Django named import DEFAULT_NAMES in migrations,
    # so we need to patch it again here.
    # Django will record `tidb_auto_id_cache` in migration files,
    # and then restore it when applying migrations.
    state.DEFAULT_NAMES += ("tidb_auto_id_cache",)


def monkey_patch():
    patch_model_functions()
    patch_model_options()
